class ChildrenPath<T> {
  ChildrenPath(
    this.getChildren,
    this.setChildren, {
    this.keepIfEmpty = false,
  });

  final dynamic Function(dynamic parent) getChildren;
  final dynamic Function(dynamic parent, dynamic children) setChildren;
  final bool keepIfEmpty;
}

class DataSearch<T> {
  DataSearch({
    required this.data,
    required this.searchMethod,
    this.depth = 0,
    this.childrenPaths = const [],
  });

  final Iterable data;
  final int depth;
  final List<ChildrenPath> childrenPaths;
  final bool Function(T dataItem) searchMethod;

  Iterable search() => _search(0, data);

  Iterable _search(int depth, Iterable data) {
    if (this.depth > depth) {
      final copyData = [];

      for (var parent in data) {
        final childrenPath = childrenPaths[depth];

        final children = childrenPath.getChildren(parent);

        final isList = children is Iterable;

        final filteredChildren =
            _search(depth + 1, isList ? children : [children]);

        if (filteredChildren.isNotEmpty || childrenPath.keepIfEmpty) {
          final copyParent = childrenPath.setChildren(
            parent,
            isList ? filteredChildren : filteredChildren.first,
          );

          copyData.add(copyParent);
        }
      }

      return copyData;
    } else {
      return data.where((element) => searchMethod(element));
    }
  }
}
